#include "stdafx.h"
#include "gl.h"
#include "GLModel.h"

const GLfloat gl0 = 0.0;
const GLfloat gl1 = 1.0;

const GLfloat PI = (GLfloat)3.14159;
const GLfloat TWOPI = (GLfloat)2.0 * PI;

const int GL_TEXTBASE = 1000;		// Lowest call list number 
const int GL_TEXTMIN = 48 ;			// Lowest drawable character A
const int GL_TEXTMAX = 122; //90;			// Highest drawable character Z

const static int GLR_SEG=24;
const static GLfloat GLR_EXT=0.333f;	// distance from center to endpoints of polygons on ribbon

//-----------------------------------------------------------------------------
CGLVector :: CGLVector( GLfloat dX, GLfloat dY, GLfloat dZ, COLORREF cr ) { 
	Set( dX, dY, dZ, cr );
	}
	
void CGLVector :: Set( GLfloat dX, GLfloat dY, GLfloat dZ, COLORREF cr ) { 
	m_dX=dX;
	m_dY=dY;
	m_dZ=dZ;
	m_cr=cr;
	}

//-----------------------------------------------------------------------------
void CGLModel :: Render(  ) { 
	for( int i=0; i<m_aObj.Count(); ++i ) 
		if( ((CGLObject *)m_aObj[i])->m_bVisible ) 
			((CGLObject *)m_aObj[i])->Render( );
	} 

void CGLModel :: SetColor( bool bColor ) { 
	for( int i=0; i<m_aObj.Count(); ++i ) 
		((CGLObject *)m_aObj[i])->m_bColor=bColor;
	} 
	
void CGLModel :: AddObject( CGLObject * pObj ) { 
	m_aObj.AddItem( pObj );
	} 

void CGLModel :: GLFontSetup( HDC hdc ) { 
	//-- create display lists for text drawing
	if( ! m_bFontSetup ) { 
		m_bFontSetup = true;
		HFONT fnt = CreateFont( 10, 0, 0, 0, FW_NORMAL, 0, 0, 0, ANSI_CHARSET, OUT_TT_PRECIS, 
								CLIP_DEFAULT_PRECIS, 
								ANTIALIASED_QUALITY /*PROOF_QUALITY*/, 
								VARIABLE_PITCH|FF_SWISS, NULL );
		HFONT fntWas = (HFONT) SelectObject( hdc, fnt );
		BOOL l1 = wglUseFontOutlines( hdc, GL_TEXTMIN, GL_TEXTMAX-GL_TEXTMIN+1, GL_TEXTBASE, 
							(GLfloat)0.1, (GLfloat)0.1, WGL_FONT_POLYGONS, NULL);
		SelectObject( hdc, fntWas );
		DeleteObject( fnt );
		}
	}

void CGLModel :: SetFine( bool bFine ) { 
	//-- set all object with fine flag to specified value (true/false) for fine grained printing.
	for( int i=0; i<m_aObj.Count(); ++i ) 
		if( GetObject(i)->m_iType==GLOB_RIBBON ) 
			((CGLRibbon *)GetObject(i))->m_bFine=bFine;
	}

CGLObject * CGLModel :: GetObject( int iObj ) { 
	return (CGLObject *)(m_aObj[iObj]);
	} 

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void CGLObject :: SetBase( GLOB_TYPE iType, GLfloat dX, GLfloat dY, GLfloat dZ,
				GLfloat dXRot, GLfloat dYRot, GLfloat dZRot,
				GLfloat dXScale, GLfloat dYScale, GLfloat dZScale ) { 
	m_dX=dX; 
	m_dY=dY; 
	m_dZ=dZ;
	m_dXRot=dXRot; 
	m_dYRot=dYRot; 
	m_dZRot=dZRot;
	m_dXScale=dXScale; 
	m_dYScale=dYScale; 
	m_dZScale=dZScale;
	m_bVisible=true;
	m_bColor=true;
	m_iType=iType;
	}

void CGLObject :: Prepare() { 
	glTranslatef( m_dX, m_dY, m_dZ );
	glRotatef( GLfloat(m_dXRot), gl0, gl1, gl0 );
	glRotatef( GLfloat(m_dYRot), gl1, gl0, gl0 );
	glRotatef( GLfloat(m_dZRot), gl0, gl0, gl1 );
	glScalef( m_dXScale, m_dYScale, m_dZScale );
	} 

void CGLObject :: Render(  ) { 
	}

void CGLObject :: RenderColor( COLORREF cr ) { 
	GLfloat f = gl1/(GLfloat)256.0;
	if( m_bColor ) 
		glColor3f( f * GetRValue(cr), f * GetGValue(cr), f * GetBValue(cr) );
	else {
		/*
		f = sqrt( f*GetRValue(cr)*f*GetRValue(cr) 
				  + f*GetGValue(cr)*f*GetGValue(cr) 
				  + f*GetBValue(cr)*f*GetBValue(cr) );
		glColor3f( 0.8f*f , 0.8f*f, 0.8f*f );		// use 80% of scale - full white is not visible
		*/
		f = GetRValue(cr)>.64 || GetGValue(cr)>.64 || GetBValue(cr)>.64 ? gl0 : gl1;
		glColor3f( f , f, f );		// use 80% of scale - full white is not visible
		}
	}

void CGLObject :: BlendColor( COLORREF cr0, COLORREF cr1, double dFrac ) { 
	RenderColor( RGB( dFrac*GetRValue(cr1) + (1-dFrac)*GetRValue(cr0), 
		 dFrac*GetGValue(cr1) + (1-dFrac)*GetGValue(cr0),
		 dFrac*GetBValue(cr1) + (1-dFrac)*GetBValue(cr0) ) );
	/*
	GLfloat f = gl1/(GLfloat)256.0;
	glColor3f( GLfloat(f*dFrac*GetRValue(cr1) + f*(1.0-dFrac)*GetRValue(cr0)),
			   GLfloat(f*dFrac*GetGValue(cr1) + f*(1.0-dFrac)*GetGValue(cr0)), 
			   GLfloat(f*dFrac*GetBValue(cr1) + f*(1.0-dFrac)*GetBValue(cr0)) );
	*/
	}


//-----------------------------------------------------------------------------
void CGLBlock :: Render2Quads( int iCR, GLfloat dx0, GLfloat dy0, GLfloat dz0,
							 GLfloat dx1, GLfloat dy1, GLfloat dz1 ) { 
							 
	RenderColor( m_cr[iCR] );
	glVertex3f( gl0, gl0, gl0 );
	glVertex3f( dx0, dy0, dz0 );
	glVertex3f( dx0+dx1, dy0+dy1, dz0+dz1 );
	glVertex3f( dx1, dy1, dz1 );
	
	RenderColor( m_cr[iCR] );
	glVertex3f( gl1, gl1, gl1 );
	glVertex3f( gl1-dx1, gl1-dy1, gl1-dz1 );
	glVertex3f( gl1-dx0-dx1, gl1-dy0-dy1, gl1-dz0-dz1 );
	glVertex3f( gl1-dx0, gl1-dy0, gl1-dz0 );
	
	}
	
void CGLBlock :: Render(  ) { 
	const GLfloat BLOCKINDENT=0.2f;
	const GLfloat BLOCKTEXTSEP=0.1f;
	
	glPushMatrix();

	if( m_bWire )
		glPolygonMode( GL_FRONT,  GL_LINE );

	Prepare();
	glBegin( GL_QUADS );
	Render2Quads( 0, gl1,gl0,gl0, gl0,gl1,gl0 );	// z=0, z=1 planes
	Render2Quads( 2, gl0,gl0,gl1, gl1,gl0,gl0 );	// y=0, y=1 planes
	Render2Quads( 4, gl0,gl1,gl0, gl0,gl0,gl1 );	// x=0, x=1 planes
	glEnd();

	if( m_bWire )
		glPolygonMode( GL_FRONT,  GL_FILL );

	if( m_pchText != NULL ) { 
		RenderColor( m_crText );
		glListBase( GL_TEXTBASE-GL_TEXTMIN);

		//-- Front face
		glPushMatrix();
		glTranslatef( BLOCKINDENT, BLOCKINDENT, 1.0f+BLOCKTEXTSEP );
		glCallLists( strlen(m_pchText), GL_UNSIGNED_BYTE, m_pchText );
		glPopMatrix();
		
		//-- Right face
		glPushMatrix();
		glTranslatef( 1.0f+BLOCKTEXTSEP, BLOCKINDENT, 1.0f-BLOCKINDENT );
		glRotatef( 90.0, 0.0, 1.0, 0.0 );
		glCallLists( strlen(m_pchText), GL_UNSIGNED_BYTE, m_pchText );
		glPopMatrix();
		
		//-- Back face
		glPushMatrix();
		glTranslatef( 1.0f-BLOCKINDENT, BLOCKINDENT, -1.0f*BLOCKTEXTSEP );
		glRotatef( 180.0, 0.0, 1.0, 0.0 );
		glCallLists( strlen(m_pchText), GL_UNSIGNED_BYTE, m_pchText );
		glPopMatrix();
		
		//-- Left face
		glPushMatrix();
		glTranslatef( -1.0f*BLOCKTEXTSEP, BLOCKINDENT, BLOCKINDENT );
		glRotatef( 270.0, 0.0, 1.0, 0.0 );
		glCallLists( strlen(m_pchText), GL_UNSIGNED_BYTE, m_pchText );
		glPopMatrix();
		
		//-- Top face
		glPushMatrix();
		glTranslatef( BLOCKINDENT, 1.0f+BLOCKTEXTSEP, 1.0f-BLOCKINDENT );
		glRotatef( -90.0, 1.0, 0.0, 0.0 );
		glCallLists( strlen(m_pchText), GL_UNSIGNED_BYTE, m_pchText );
		glPopMatrix();

		//-- Bottom face
		glPushMatrix();
		glTranslatef( BLOCKINDENT, -1.0f*BLOCKTEXTSEP, BLOCKINDENT );
		glRotatef( 90.0, 1.0, 0.0, 0.0 );
		glCallLists( strlen(m_pchText), GL_UNSIGNED_BYTE, m_pchText );
		glPopMatrix();
		}

	glPopMatrix();
	}
	
void CGLBlock :: SetColors( COLORREF crZ0, COLORREF crZ1, COLORREF crX0, COLORREF crX1, COLORREF crY0, COLORREF crY1 ) { 
	m_cr[0]=crZ0;
	m_cr[1]=crZ1;
	m_cr[2]=crX0;
	m_cr[3]=crX1;
	m_cr[4]=crY0;
	m_cr[5]=crY1;
	}

void CGLBlock :: SetText( COLORREF cr, const char * pchText ) { 
	m_pchText=new char[strlen(pchText)+1];
	strcpy( m_pchText, pchText );
	m_crText = cr;
	}

CGLBlock :: CGLBlock() { 
	m_pchText=NULL;
	m_bWire=false;
	m_iType=GLOB_BLOCK;
	}
	
CGLBlock :: ~CGLBlock() { 
	if( m_pchText!=NULL )
		delete [] m_pchText;
	}

//-----------------------------------------------------------------------------
void CGLPolyCyl :: Set( int iFaces, COLORREF cr0, COLORREF cr1, COLORREF crBody ) { 
	m_iFaces=max(2,min(48,iFaces));
	m_cr0=cr0;
	m_cr1=cr1;
	m_crBody=crBody;
	m_iType=GLOB_POLYCYL;
	}
	
void CGLPolyCyl :: Render(  ) { 
	double nAngle = TWOPI / m_iFaces;

	//--			
	glPushMatrix();
	Prepare();

	//-- x=0 end cap 
	RenderColor( m_cr0 );
	glBegin( GL_POLYGON );
	for( int i=0; i<m_iFaces; ++i ) 
		glVertex3f( gl0, GLfloat( cos( nAngle * i ) ), GLfloat( sin( nAngle * i ) ) );
	glEnd();
	
	//-- x=1 end cap 
	RenderColor( m_cr0 );
	glBegin( GL_POLYGON );
	for( int i=0; i<m_iFaces; ++i ) 
		glVertex3f( gl1, GLfloat(cos( -1 * nAngle * i )), GLfloat(sin( -1 * nAngle * i ) ) );
	glEnd();
	
	//-- quad strip for body 
	RenderColor( m_crBody );
	glBegin( GL_QUAD_STRIP );
	glVertex3f( gl0, gl1, gl0 );
	glVertex3f( gl1, gl1, gl0 );
	for( int i=1; i<m_iFaces; ++i ) { 
		glVertex3f( gl0, GLfloat(cos( nAngle * i )), GLfloat(sin( nAngle * i )) );
		glVertex3f( gl1, GLfloat(cos( nAngle * i )), GLfloat(sin( nAngle * i )) );
		} 
	glVertex3f( gl0, gl1, gl0 );
	glVertex3f( gl1, gl1, gl0 );
	glEnd();
	
	//-- 
	glPopMatrix();
	}
	
//-----------------------------------------------------------------------------
void CGLRibbon :: Set( int iFaces, bool bEndCaps,
					GLfloat dX0, GLfloat dY0, GLfloat dZ0,
					GLfloat dYm,
					GLfloat dX1, GLfloat dY1, GLfloat dZ1, 
					GLfloat dWidthFactor,
					COLORREF cr0, COLORREF crM, COLORREF cr1 ) {
	m_bEndCaps = bEndCaps;					
	m_iFaces=iFaces;
	m_dYm=dYm-dY0;				  
	m_dY1=dY1-dY0;			
	m_dWidthFactor=max( GLfloat(0.0), min(GLfloat(1.0),dWidthFactor) ) ;	 
	double dXX = (dX1-dX0); 
	double dZZ = (dZ1-dZ0); 
	double dH = sqrt( dXX*dXX+dZZ*dZZ );
	m_dLen = GLfloat( sqrt( dXX*dXX+dZZ*dZZ ) );
	SetBase( GLOB_RIBBON, dX0, dY0, dZ0, gl0, gl0, gl0, gl1, gl1, gl1 );
	if( dXX==0.0 )
		m_dXRot = dZZ>0 ? PI*(GLfloat)0.5 : PI*(GLfloat)1.5;
	else {
		m_dXRot = GLfloat( asin(dZZ/dH) );
		if( dXX<0.0 ) 
			m_dXRot = PI - m_dXRot ;
		}
	m_dXRot = m_dXRot * -180.0f /PI;
	m_cr0=cr0;
	m_crM=crM;
	m_cr1=cr1;
	m_bFine=false;
	m_iType=GLOB_RIBBON;
	}

void CGLRibbon :: Render( ) { 
	//-- 
	glPushMatrix();
	Prepare();

	int i, iFace;
	double dFrac;
	GLfloat dSin, dCos;
	double d0, d1;					// un rotated angle to start and end of face 
	GLfloat dXc, dYc, dZc;			// center or ribbon section being drawn
	GLfloat dXe,dYe,dZe,dXYe;

	if( m_bFine ) { 	
		glEnable(GL_LINE_SMOOTH);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
		//glLineWidth(1.5);
		glDisable(GL_DEPTH_TEST);		// to correct line segment endpoint issue

		//-- quad strip for 1st half of body 
		glBegin( GL_LINE_STRIP );
		for( i=0; i<=GLR_SEG; ++i ) { 
			dCos = cos( PI * i / GLR_SEG );	// height weighting is a cosine function
			dCos = 0.5f * ( dCos+1.0f );
			dSin = sin( PI * i / GLR_SEG );	// rotation angle is -1/ ( dCos/dt ) = -1 / Sin
			
			dFrac = double(i) / double(GLR_SEG);// fraction of planar distance covered
			
			//-- set vertex color as mix of here and there 
			BlendColor( m_cr0, m_crM, dFrac );

			//-- calc center of next pos 
			dXc = (GLfloat)0.5 * m_dLen * (GLfloat)dFrac * GLfloat(1.0-m_dWidthFactor);
			dYc = m_dYm*(1-dCos);
			dZc = gl0;
			glVertex3f( dXc, dYc, dZc );
			}
		
		//-- quad strip for 2nd half of body 
		for( i=GLR_SEG; i<=GLR_SEG*2; ++i ) { 
			dCos = cos( PI * i / GLR_SEG );	// height weighting is a cosine function
			dCos = 0.5f * ( dCos+1.0f );
			dSin = sin( PI * i / GLR_SEG );	// rotation angle is -1/ ( dCos/dt ) = -1 / Sin
			dFrac = double(i-GLR_SEG) / double(GLR_SEG);// fraction of planar distance covered

			//-- set vertex color as mix of here and there 
			BlendColor( m_crM, m_cr1, dFrac );

			//-- calc center of next pos 
			// dXc = m_dLen * 0.5f * (gl1 +(GLfloat)dFrac); 
			// Changed to add plateau in middle 'WidthFactor'
			dXc = GLfloat( m_dLen * ( (gl1+m_dWidthFactor)/2.0 + 0.5f *(GLfloat)dFrac* GLfloat(1.0-m_dWidthFactor) ) );
			dYc = m_dY1*dCos + m_dYm*(1-dCos);
			dZc = gl0 ;
			glVertex3f( dXc, dYc, dZc );
			}
			glEnd();
		glEnable(GL_DEPTH_TEST);
		}
	else { 
		//-- quad strip for 1st half of body 
		for( iFace=0; iFace<m_iFaces; iFace++ ) { 
			d1 = TWOPI * iFace / m_iFaces;
			d0 = TWOPI * (iFace+1) / m_iFaces;
			
			glBegin( GL_QUAD_STRIP );
			for( i=0; i<=GLR_SEG; ++i ) { 
				dCos = cos( PI * i / GLR_SEG );	// height weighting is a cosine function
				dCos = 0.5f * ( dCos+1.0f );
				dSin = sin( PI * i / GLR_SEG );	// rotation angle is -1/ ( dCos/dt ) = -1 / Sin
				
				dFrac = double(i) / double(GLR_SEG);// fraction of planar distance covered
				
				//-- set vertex color as mix of here and there 
				BlendColor( m_cr0, m_crM, dFrac );

				//-- calc center of next pos 
				dXc = (GLfloat)0.5 * m_dLen * (GLfloat)dFrac * GLfloat(1.0-m_dWidthFactor);
				dYc = m_dYm*(1-dCos);
				dZc = gl0;
				
				//-- plot 
				dZe = GLR_EXT*GLfloat(cos(d0));
				dXYe = GLfloat(sqrt( GLR_EXT*GLR_EXT - dZe*dZe ) * ( d0>0.5*PI && d0<1.5*PI ? 1.0f : -1.0f ));
				dYe = GLfloat(sqrt( max(0.0,dXYe*dXYe/(1+dSin*dSin)) ) * (  d0>PI ? 1.0f : -1.0f )) ;
				dXe = dYe * dSin;
				glVertex3f( dXc-dXe, dYc+dYe, dZc+dZe );

				dZe = GLR_EXT*GLfloat(cos(d1));
				dXYe = GLfloat(sqrt( GLR_EXT*GLR_EXT - dZe*dZe ) * ( d1>0.5*PI && d1<1.5*PI ? 1.0f : -1.0f ));
				dYe = GLfloat(sqrt( max(0.0,dXYe*dXYe/(1+dSin*dSin)) ) * (  d1>PI ? 1.0f : -1.0f )) ;
				dXe = dYe * dSin;
				glVertex3f( dXc-dXe, dYc+dYe, dZc+dZe );
				}
			
			//-- quad strip for 2nd half of body 
			for( i=GLR_SEG; i<=GLR_SEG*2; ++i ) { 
				dCos = cos( PI * i / GLR_SEG );	// height weighting is a cosine function
				dCos = 0.5f * ( dCos+1.0f );
				dSin = sin( PI * i / GLR_SEG );	// rotation angle is -1/ ( dCos/dt ) = -1 / Sin
				dFrac = double(i-GLR_SEG) / double(GLR_SEG);// fraction of planar distance covered

				//-- set vertex color as mix of here and there 
				BlendColor( m_crM, m_cr1, dFrac );

				//-- calc center of next pos 
				// dXc = m_dLen * 0.5f * (gl1 +(GLfloat)dFrac); 
				// Changed to add plateau in middle 'WidthFactor'
				dXc = GLfloat( m_dLen * ( (gl1+m_dWidthFactor)/2.0 + 0.5f *(GLfloat)dFrac* GLfloat(1.0-m_dWidthFactor) ) );
				dYc = m_dY1*dCos + m_dYm*(1-dCos);
				dZc = gl0 ;
				
				//-- plot 
				dZe = GLR_EXT*GLfloat(cos(d0));
				dXYe = GLfloat(sqrt( GLR_EXT*GLR_EXT - dZe*dZe ) * ( d0>0.5*PI && d0<1.5*PI ? 1.0f : -1.0f ));
				dYe = GLfloat(sqrt( max(0.0,dXYe*dXYe/(1+dSin*dSin)) ) * (  d0>PI ? 1.0f : -1.0f )) ;
				dXe = dYe * dSin;
				glVertex3f( dXc-dXe, dYc+dYe, dZc+dZe );
				
				dZe = GLR_EXT*GLfloat(cos(d1));
				dXYe = sqrt( GLR_EXT*GLR_EXT - dZe*dZe ) * ( d1>0.5*PI && d1<1.5*PI ? 1.0f : -1.0f );
				dYe = GLfloat(sqrt( max(0.0,dXYe*dXYe/(1.0f+dSin*dSin)) ) * (  d1>PI ? 1.0f : -1.0f ));
				dXe = dYe * dSin;
				glVertex3f( dXc-dXe, dYc+dYe, dZc+dZe );
				}
				glEnd();
			}
		}

	if( ! m_bFine ) { 			
		//----- First end cap (Polygon)
		RenderColor( m_cr0 );
		glBegin( GL_POLYGON );
		//-- draw polygon
		for( iFace=0; iFace<m_iFaces; ++iFace ) { 
			d0 = TWOPI * iFace / m_iFaces;
			dZe = GLR_EXT*GLfloat(cos(d0));
			dYe = sqrt( GLR_EXT*GLR_EXT - dZe*dZe ) * ( d0>PI ? 1.0f : -1.0f );
			glVertex3f( 0.0, dYe, dZe );
			}
		glEnd();
			
		//----- Second end cap (Polygon)
		RenderColor( m_cr1 );
		glBegin( GL_POLYGON );
		for( iFace=m_iFaces-1; iFace>=0; --iFace ) { 
			d0 = TWOPI * iFace / m_iFaces;
			dZe = GLR_EXT*GLfloat(cos(d0));
			dYe = sqrt( GLR_EXT*GLR_EXT - dZe*dZe ) * ( d0>PI ? 1.0f : -1.0f );
			glVertex3f( m_dLen, m_dY1+dYe, dZe );
			}
		glEnd();
		}
		
	//-- 
	glPopMatrix();
	}

//-----------------------------------------------------------------------------
CGLText :: CGLText(){
	m_pchText=NULL;
	m_iType=GLOB_TEXT;
	}

CGLText :: ~CGLText(){
	if( m_pchText != NULL ) 
		delete [] m_pchText;	
	}

void CGLText :: Set( COLORREF cr, const char * pchText ){
	m_cr=cr;
	if( m_pchText != NULL ) 
		delete [] m_pchText;
	if( (m_pchText = new char[strlen(pchText)+1]) != NULL ) 
		strcpy(m_pchText,pchText);
	}

void CGLText :: Render(  ) { 
	glPushMatrix();
	
	Prepare();
	RenderColor( m_cr );
    glListBase( GL_TEXTBASE-GL_TEXTMIN);
	glCallLists( strlen(m_pchText), GL_UNSIGNED_BYTE, m_pchText );
	
	glPopMatrix();
	}

	
	
	








